/*
   Aclan, Nelson D.
   simulationelson@gmail.com
   Last Modified: September 18, 2009

   A (closed) hash table whose cells can contain two elements;
   makes use of a rehashing function to constantly search
   for an available cell when a collision occurs.
*/

#include <stdio.h>
#define MAX_VALUE 13

typedef struct CHash {
   int numElements;
   int cell[MAX_VALUE][2];
} CHash;

void initCHash(CHash *ch);
int isEmptyCH(CHash *ch);
int isFullCH(CHash *ch);
void insert_C(CHash *ch, int key);
void rehashAndHush(CHash *ch, int hash_second, int *rehash_index);
void search_C(CHash *ch, int key);
void printCHash(CHash *ch);
void emptyCHash(CHash *ch);
void underflow();

main() {
   CHash ch;
   initCHash(&ch);

   char choice[2];
   int input;

   do {
      printf("-----------------------------------\n");
      printf("|    Closed Hash Menu             |\n");
      printf("-----------------------------------\n");
      printf("|    (1) Insert Key               |\n");
      printf("|    (2) Search Key               |\n");
      printf("|    (3) Display All Keys         |\n");
      printf("|    (4) Delete All Keys          |\n");
      printf("|    (5) Exit                     |\n");
      printf("-----------------------------------\n");

      printf("Enter choice: ");
      scanf("%s", choice);

      switch(choice[0]) {
         case '1': /* Insert Key */
            if(isFullCH(&ch)) {
               printf("Hash Table Full\n");
            } else {
               printf("Enter key to be inserted: ");
               scanf("%d", &input);

               if(input <= 0) printf("Invalid key\n");
               else insert_C(&ch, input);
            }
            break;
         case '2': /* Search Key */
            if(isEmptyCH(&ch)) {
               underflow();
            } else {
               printf("Enter key to be searched: ");
               scanf("%d", &input);

               if(input <= 0) printf("Invalid key\n");
               else search_C(&ch, input);
            }
            break;
         case '3': /* Display All Keys */
            printCHash(&ch);
            break;
         case '4': /* Delete All Keys */
            emptyCHash(&ch);
            break;
         case '5': /* Exit */
            return;
            break;
         default: /* None of the above */
            printf("Bad move\n");
            break;
         }
   } while(choice[0] != '5');
}

/* Initialize closed hash table */
void initCHash(CHash *ch) {
   int i, j;

   ch->numElements = 0;

   for(i = 0; i < MAX_VALUE; i++) {
      for(j = 0; j < 2; j++) {
         ch->cell[i][j] = 0;
      }
   }
}

/* Check if hash table is empty */
int isEmptyCH(CHash *ch) {
   if(ch->numElements == 0) {
      return 1;
   } else {
      return 0;
   }
}

/* Check if hash table is full */
int isFullCH(CHash *ch) {
   if(ch->numElements == MAX_VALUE*2) {
      return 1;
   } else {
      return 0;
   }
}

/* Insert a key into the hash table */
void insert_C(CHash *ch, int key) {
   int hash_index, hash_second, rehash_index;

   hash_index = key % MAX_VALUE; /* h(k) = i = k mod 13; (Hashing Function) */
   hash_second = key % (MAX_VALUE - 2); /* h'(k) = k mod 11; */
   rehash_index = (hash_second + hash_index) % MAX_VALUE; /* h(k,i) = (h'(k) + i) mod 13; (Rehashing Function) */

   /* start of hash_index availability */
   if(ch->cell[hash_index][0] == 0) {
      ch->cell[hash_index][0] = key;
   } else if((ch->cell[hash_index][0] != 0) && (ch->cell[hash_index][1] == 0)) {
      ch->cell[hash_index][1] = key;
   /* end of hash_index availability */
   } else {
      /* start of rehash_index availability */
      if(ch->cell[rehash_index][0] == 0) {
         ch->cell[rehash_index][0] = key;
      } else if((ch->cell[rehash_index][0] != 0) && (ch->cell[rehash_index][1] == 0)) {
         ch->cell[rehash_index][1] = key;
      /* end of rehash_index availability */
      } else {
         rehashAndHush(ch, hash_second, &rehash_index); /* search an available cell */

         /* start of NEW rehash_index availability */
         if(ch->cell[rehash_index][0] == 0) {
            ch->cell[rehash_index][0] = key;
         } else if((ch->cell[rehash_index][0] != 0) && (ch->cell[rehash_index][1] == 0)) {
            ch->cell[rehash_index][1] = key;
         /* end of NEW rehash_index availability */
         } else {
            /* Take a break */
         }
      }
   }

   printf("Key (%d) inserted\n", key);
   ch->numElements++;
}

/* Repeatedly rehash in search for an available cell */
void rehashAndHush(CHash *ch, int hash_second, int *rehash_index) {
   *rehash_index = (hash_second + *rehash_index) % MAX_VALUE;

   if((ch->cell[*rehash_index][0] != 0) && (ch->cell[*rehash_index][1] != 0)) {
      rehashAndHush(ch, hash_second, rehash_index);
   }
}

/* Search a key in the hash table */
void search_C(CHash *ch, int key) {
   int i, j, found;

   found = 0;

   for(i = 0; i < MAX_VALUE; i++) { /* does not use any of the hashing functions...why? */
      for(j = 0; j < 2; j++) { /* because the same key could be located elsewhere */
         if(ch->cell[i][j] == key) {
            printf("Key (%d) is found in hash table cell no. %d element %d\n", key, i, j+1);
            found++;
         }
      }
   }

   if(!found) {
      printf("Key not found\n");
   }
}

/* Print hash table contents */
void printCHash(CHash *ch) {
   if(isEmptyCH(ch)) {
      underflow();
   } else {
      int i;

      for(i = 0; i < MAX_VALUE; i++) {
         if((ch->cell[i][0] != 0) && (ch->cell[i][1] == 0)) { /* first cell is occupied */
            printf("Cell[%d] = {%d}\n", i, ch->cell[i][0]);
         } else if((ch->cell[i][0] == 0) && (ch->cell[i][1] != 0)) { /* second cell is occupied */
            printf("Cell[%d] = {%d}\n", i, ch->cell[i][1]);
         } else if((ch->cell[i][0] != 0) && (ch->cell[i][1] != 0)) { /* both cells are occupied  */
            printf("Cell[%d] = {%d,%d}\n", i, ch->cell[i][0], ch->cell[i][1]);
         } else { /* both cells are not occupied */
            /* Print nothingness */
         }
      }
   }
}

/* Empty the hash table */
void emptyCHash(CHash *ch) {
   if(isEmptyCH(ch)) {
      underflow();
   } else {
      initCHash(ch);
      printf("Hash Table Emptied\n");
   }
}

/* Nothing much */
void underflow() {
   printf("Hash Table Empty\n");
}
